import typing

from pg_sql import SqlObject, sql_list

from .format import format
from .formats.join import JoinRefresh, JoinTargetTable
from .join_common import JoinTarget, Key
from .sql import SqlQuery
from .sql_query import insert_query, sync_query, upsert_query


class JoinTableTarget(JoinTarget):
    """
    Insert into to a table
    """

    def __init__(self, table: JoinTargetTable, query: str):
        self._table = table
        self._query = query

    def key(self) -> typing.Optional[Key]:
        if self._table.key:
            definition = f"""
SELECT {sql_list(self._table.key)}
FROM {self._table.sql}
            """.strip()
            names = self._table.key
            return Key(definition=definition, names=names)

    def sql(self, key_table: SqlObject) -> SqlQuery:
        formatted = format(self._query, str(key_table))

        if self._table.refresh == JoinRefresh.FULL:
            query = sync_query(
                columns=self._table.columns or self._table.key,
                key=self._table.key,
                key_table=key_table,
                query=formatted,
                target=self._table.sql,
            )
        elif self._table.refresh == JoinRefresh.INSERT:
            query = insert_query(
                columns=self._table.columns or self._table.key,
                query=formatted,
                target=self._table.sql,
            )
        elif self._table.refresh == JoinRefresh.UPSERT:
            query = upsert_query(
                columns=self._table.columns or self._table.key,
                key=self._table.key,
                query=formatted,
                target=self._table.sql,
            )

        return query



def format(format, *args):
    i = 0
    state = "normal"
    replacement = ""
    result = ""
    while True:
        if state == "normal":
            if len(format) <= i:
                break
            elif format[i] == "$":
                state = "replacement"
                i += 1
            else:
                result += format[i]
                i += 1
        elif state == "replacement":
            if len(format) <= i:
                n = int(replacement) - 1
                if not (0 <= n < len(args)):
                    raise RuntimeError(f"Invalid param '{replacement}' in '{format}'")
                result += args[n]
                replacement = ""
                state = "normal"
            elif format[i] == "$":
                result += "$"
                state = "normal"
                i += 1
            elif format[i].isdigit():
                replacement += format[i]
                i += 1
            else:
                n = int(replacement) - 1
                if not (0 <= n < len(args)):
                    raise RuntimeError(f"Invalid param '{replacement}' in '{format}'")
                result += args[n]
                replacement = ""
                state = "normal"

    return result



from pg_sql import SqlString

from .formats.join import JoinTable
from .join_async import enqueue_sql
from .join_common import Structure
from .string import indent


def create_refresh_function(structure: Structure, table_id: str, table: JoinTable):
    enqueue = enqueue_sql(table_id, table, structure, "SELECT false AS _", [], None)

    function = structure.refresh_table_function(table_id)

    yield f"""
CREATE FUNCTION {function}() RETURNS void
LANGUAGE plpgsql AS $$
  BEGIN
{indent(enqueue, 2)}
  END;
$$
    """.strip()

    yield f"""
COMMENT ON FUNCTION {function} IS {SqlString(f"Recalculate all, based on {table.join}")}
"""
